#include <stdio.h>

FILE * infile ;
int nt , nc ;
#define MAXTAX   10 
#define MAXCHAR  150
#define MAXNOD   20 
#define MAXSWAPS 380 // This is the max number of TBR rearrangs (on 10 taxa)
#define MAXTREES 100000

char taxname[MAXTAX] ;
char tmpname[50] ; 
unsigned char matrix[MAXNOD][MAXCHAR] ; 

int testgroup[25] , testgroupsz[25] ;
int numtestgroups = 0 ;
unsigned long int testmat[MAXNOD] ; 

int mptlen = 10000000 ;
int lptlen = 0 ;
int stage ;
int asisonly = 0 ;
int upinserts = 0 , downinserts = 0 ; 

double daminprob = 10 , damaxprob = 0 ; 

typedef struct {
    char tree [ MAXNOD ] ;
    unsigned long int key ;
    unsigned short int lokey ; 
    int len ;
    void * to ;
    double prob ; } Treetyp ; 
Treetyp * trees ;
Treetyp ** treelist ; 
int ntrees = 0 , nodz ;
char anx[MAXNOD] , sis[MAXNOD] , lef[MAXNOD] , rig[MAXNOD] , nodfork[MAXNOD] ;
char hsis[MAXNOD] , hlef[MAXNOD] , hrig[MAXNOD] ;
char tmptree [MAXNOD] ; 
char permlist[ MAXTAX ] ; 
char * anxpt ;
unsigned long int curkey ;
unsigned short int curlokey ;

unsigned long int nummpts = 0 , totperms , numperms = 0 ;
double curprobis ; 

int use_spr = 0 ;

char testpartit[25][10] , partitsize[25] , partitat[25] ;
int numtestpartits = 0 ;

void addtestgroup ( char * txt )
{
    int a ;
    char * cp = txt ;
    int slashisat = -1 , isapartit = 0 ;
    while ( * cp ) if ( * cp ++ == '/' ) isapartit = 1 ; 
    if ( isapartit ) {
       partitsize [ numtestpartits ] = partitat [ numtestpartits ] = 0 ; 
       cp = txt - 1 ;
       isapartit = 0 ; 
       while ( * ++ cp ) {
           if ( * cp == '/' ) {
               if ( partitsize [ numtestpartits ] < 2 ) break ;
               partitat [ numtestpartits ] = slashisat = partitsize [ numtestpartits ] ;
               continue ; }
           for ( a = 0 ; a < nt ; ++ a )
             if ( tolower ( taxname[a] ) == tolower ( * cp ) ) 
                testpartit [ numtestpartits ] [ partitsize [ numtestpartits ] ++ ] = a ; }
       if ( partitsize [ numtestpartits ] < 4 ||  partitsize [ numtestpartits ] - slashisat < 2 ) {
           fprintf ( stderr , "\nOOPS!! wrong partition definition: %s \nDefinition must include 2 taxa (or more) on each side.\n" , txt ) ;
           exit ( 0 ) ; }
       ++ numtestpartits ; }
    else { 
       cp = txt - 1 ;
       testgroup [ numtestgroups ] = testgroupsz [ numtestgroups ] = 0 ;
       while ( * ++ cp ) 
           for ( a = 0 ; a < nt ; ++ a )
             if ( tolower ( taxname[a] ) == tolower ( * cp ) ) {
                ++ testgroupsz [ numtestgroups ] ; 
                if ( testgroupsz [ numtestgroups ] == nt - 1 ) {
                    fprintf ( stderr , "\nOOPS!! Error in group definition: %s\nDefinition can include at most T-2 taxa.\n" , cp ) ;
                    exit ( 0 ) ; }
                testgroup [ numtestgroups ] |= 1 << a ; }
       ++ numtestgroups ; } 
}    

void readdata ( char * nam )
{
    char * nampt ;
    char state ;
    int a , b ; 
    infile = fopen ( nam , "rb" ) ;
    if ( infile == NULL ) {
        fprintf ( stderr , "Can't open %s" , nam ) ;
        exit ( 0 ) ; }
    fscanf ( infile , " %s" , tmpname ) ;
    if ( strcmp ( tmpname , "xread" ) ) {
        fprintf ( stderr , "String \"xread\" must appear first in data set" ) ;
        exit ( 0 ) ; }
    fscanf ( infile , " %i %i" , &nc , &nt ) ;
    if ( nt > MAXTAX ) {
        fprintf ( stderr , "Cannot run more than %i taxa" , MAXTAX ) ;
        exit ( 0 ) ; }
    if ( nc > MAXCHAR) {
        fprintf ( stderr , "Cannot run more than %i characters" , MAXCHAR ) ;
        exit ( 0 ) ; }
    for ( a = 0 ; a < nt ; ++ a ) {
        nampt = tmpname ; // 
        fscanf ( infile , " %s" , nampt ) ;
        if ( strlen ( nampt ) > 1 ) {
            fprintf ( stderr , "Taxon names must be single letters!" ) ;
            exit ( 0 ) ; }
        taxname[a] = * nampt ; 
        for ( b = 0 ; b < nc ; ++ b ) {
            fscanf ( infile , " %c" , &state ) ;
            if ( state == '?' || state == '-' ) matrix[a][b] = 255 ;
            else {
                if ( toupper ( state ) == 'A' || 
                    toupper ( state ) == 'G' || 
                     toupper ( state ) == 'C' || 
                     toupper ( state ) == 'T' ) {
                  if ( toupper(state) == 'A' ) matrix[a][b] = 1 ;
                  else if ( toupper(state) == 'G' ) matrix[a][b] = 2 ;
                  else if ( toupper(state) == 'C' ) matrix[a][b] = 4 ;
                  else if ( toupper(state) == 'T' ) matrix[a][b] = 8 ; }
                  else {
                    if ( state > '7' ) {
                   fprintf ( stderr , "Cannot run more than 8 states (0-7)" ) ;
                   exit ( 0 ) ; }
                   matrix[a][b] = 1 << ( state - '0' ) ; }}}}
    fclose ( infile ) ; 
    trees = ( Treetyp * ) malloc ( MAXTREES * sizeof ( Treetyp ) ) ;
    if ( trees == NULL ) {
        fprintf ( stderr , "\nOOPS...\nCannot allocate needed memory\n" ) ;
        exit ( 0 ) ; }
    treelist = ( Treetyp ** ) malloc ( 65536 * sizeof ( Treetyp * ) ) ;
    if ( treelist == NULL ) {
        fprintf ( stderr , "\nOOPS...\nCannot allocate needed memory\n" ) ;
        exit ( 0 ) ; }
    for ( a = 0 ; a < 65535 ; ++ a ) treelist [ a ] = NULL ;
    nodz = 2*nt - 1 ; 
    return ;
}        

char * bchlist , * bast ;
static int listhas, htuz, agre, totax;

void hashmksis ( char * an )
{ signed char list [ MAXNOD ] ; 
  signed char * x = list , * y = list + 1;
  signed char * last , nd ; 
  int a , i;
  last = list ; 
  for ( a = nodz ; a -- ; ) 
      hsis [ a ] = hlef [ a ] = list [ a ] = -1 ;
  for ( a = nt ; a -- ; ) {
      i = an [ nd = a ] ; 
      if ( !i ) continue ; 
      while ( i != nt && hlef [ i ] < 0 ) {
                    last [ i ] = hlef [ i ] = nd ;
                    nd = i ;
                    i = an [ i ] ; }
      if ( hlef [ i ] < 0 ) hlef [ i ] = nd ; // i.e. i == nt
      if ( last [ i ] >= 0 ) hsis [ last [ i ] ] = nd ;
      last [ i ] = nd ; }
   return ;
} 

int mkoptlist ( char *list, char *an )  
{ signed char * x = list ; 
  signed char * y = list + 1;
  signed char * last , nd ; 
  int a, i;
  last = list ; 
  for ( a = nodz ; a -- ; ) {
      sis [ a ] = lef [ a ] = list [ a ] = -1 ;
      nodfork [ a ] = 0 ; } 
  for ( a = nt ; a -- ; ) {
      i = an [ nd = a ] ; 
      if ( !i ) continue ; 
      while ( i != nt && lef [ i ] < 0 ) {
                    last [ i ] = lef [ i ] = nd ;
                    ++ nodfork [ i ] ; 
                    nd = i ;
                    i = an [ i ] ; }
      ++ nodfork [ i ] ;
      if ( lef [ i ] < 0 ) lef [ i ] = nd ; // i.e. i == nt
      if ( last [ i ] >= 0 ) sis [ last [ i ] ] = nd ;
      last [ i ] = nd ; }
   * x = nt ; 
   while ( x < y ) {
       a = lef [ * x ++ ] ;
       while ( a >= 0 ) { if ( a >= nt ) * y ++ = a ; a = sis [ a ] ; }}
   return ( x - ( signed char *) list ) ;
} 

void reset_nodz ( char * an )
{
  char bu [ MAXNOD ] ; 
  char tbu [ MAXNOD ] ;
  int a, b, atno, lg ; 
  for ( a = nodz ; a-- > nt ; ) tbu [ a ] = -1;
  for ( a = nt ; a-- ; ) tbu [ a ] = a ; 
  atno = nt ;
  lg = an [ nt ] ; 
  tbu [ nt ] = nt ;
  tbu [ lg ] = lg ; 
  for ( a = 0 ; a < nt ; ++a) {
      b = an [ a ] ;
      if ( !b ) continue ; 
      while ( tbu [ b ] < 0 ) {
          tbu [ b ] = ++ atno ; 
          b = an [ b ] ; }}
  for ( a = lg ; a-- ; ) 
      bu [ tbu [ a ] ] = tbu [ an [ a ] ] ;
  for ( a = lg ; a-- ; ) an [ a ] = bu [ a ] ;
  an [ nt ] = ++ atno ; 
  return ; 
}

unsigned long int topnod; 
short int bchbel [ MAXNOD ] ;
long int needbitz ;

static unsigned long int bm[] = {
       0,1,3,7,15,31,63,127,255,511,1023,2047,4095,8201 } ;
       
void hashtree ( char * treept )     // Make a unique 32-bit tree-identifier... 
{ unsigned long int a, b, n, jamp, at, low;
  unsigned long int bitz, value, ovr, anxwas; 
  hashmksis( treept ) ;
  for( a = 1 ; a < nodz ; ++ a ) bchbel [ a ] = -1 ;
  bchbel [ a = 0 ] = 0;
  while( a != nt ) { a = treept [ a ] ; bchbel [ a ] = 0 ; }
  a = 1 ; while ( bchbel [ a ] < 0 ) { bchbel [ a ] = 1 ; a = treept [ a ] ; }
  bchbel [ a ] = 2 ; while( a != nt ) { bchbel [ a ] = 2 ; a = treept [ a ] ; }
  bchbel [ nt ] = jamp = bitz = 2;
  needbitz = 0; 
  for ( a = nt ; a < nodz ; ++ a ) hrig [ a ] = 0 ;
  a = 0 ; while ( a != nt ) { a = treept [ a ] ; ++ hrig [ a ] ; }
  a = treept [ 1 ] ; while( ! hrig [ a ] ) { ++ hrig [ a ] ; a = treept [ a ] ; }
  ++ hrig [ a ] ; 
  curkey = 0 ; 
  ovr = 0 ;
  treept [ nt ] = 0 ; 
  for( a = 2 ; a < nt ; ++ a ) { 
     n = a ; while ( bchbel [ n ] < 0 ) n = treept [ n ] ;
     while( ( jamp & bm [ bitz ] ) != jamp ) ++ bitz ;
     needbitz += ( bitz + 1 ) ; 
     if( ++ hrig [ n ] > 2 ) { 
         low = n ; 
         n = hlef [ n ] ;
         while ( bchbel [ n ] < 0 ) n = hsis [ n ] ;
         value = ( bchbel [ n ] << 1 ) | 1 ;
         n = treept [ a ] ;
         while( ! hrig [ n ] ) { ++ hrig [ n ] ; n = treept [ n ] ; }   
         n = a;
         ++ jamp; 
         while( n != low ) { bchbel [ n ] = jamp ; n = treept [ n ] ; } 
         bchbel [ a ] = jamp ; }
     else {
         at = bchbel [ n ] ;
         b = low = n ;
         while ( bchbel [ low ] == at ) low = treept [ low ] ;
         n = hlef [ n ] ; while ( bchbel [ n ] < 0 ) n = hsis [ n ] ;  
         value = bchbel [ n ] << 1 ; 
         n = treept [ a ] ; while ( ! hrig [ n ] ) { ++ hrig [ n ] ; n = treept [ n ] ; } 
         ++ jamp; 
         n = a ; while( n != b ) { bchbel [ n ] = jamp ; n = treept [ n ] ; } 
         ++ jamp;
         while( n != low ) { bchbel [ n ] = jamp ; n = treept [ n ] ; }}
     curkey |= ( value << ovr );
     ovr += bitz + 1 ; }
  curlokey =
      ( curkey & 65535 ) ^
      ( ( curkey & ( 65535 << 16 ) ) >> 16 ) ;
  return; 
}

void zert ( int whatp, int wherep )
{
    int a, b, c ;
    if ( wherep == 0 ) {
        anxpt [ lef [ nt ] = sis [ whatp ] = 0 ] = nt ; 
        anxpt [ rig [ nt ] = sis [ 0 ] = whatp ] = nt ;
        return ; } 
    a = anxpt [ wherep ] ;
    c = sis [ wherep ] ; 
    b = anxpt [ wherep ] = anxpt [ whatp ] ; 
    anxpt [ b ] = a;
    sis [ sis [ wherep ] = rig [ b ] = whatp ] = lef [ b ] = wherep;
    sis [ lef [ a ] = c ] = b ;
    sis [ rig [ a ] = b ] = c ;
}     

void unzert ( int whatp )
{  
    int a, b, c;
    a = anxpt [ whatp ] ;
    if ( a == nt ) {
        lef [ anxpt [ 0 ] = nt ] = rig [ nt ] = sis [ 0 ] = 0 ;
        anxpt [ whatp ] = whatp ;
        return ; } 
    b = sis [ whatp ] ;
    c = sis [ a ] ; 
    a = anxpt [ a ] ; 
    lef [ a ] = b ;
    rig [ anxpt [ b ] = anxpt [ c ] = a ] = sis [ sis [ c ] = b ] = c;
}

int binoptlist ( char * list , char * tt )
{
    char * x , * y ;
    x = y = list ; 
    * y ++ = nt ;
    while ( x < y ) {
        if ( lef [ * x ] >= nt ) * y ++ = lef [ * x ] ;
        if ( rig [ * x ] >= nt ) * y ++ = rig [ * x ] ;
        ++ x ; }
    return y - list ;
}

void binmksis ( char * tt ) 
{
    int a , b , c , lownod = tt [ nt ] ;
    char lassis [ MAXNOD ] ; 
    for ( a = nodz ; a -- ; ) nodfork [ a ] = 0 ;
    for ( a = nt ; a -- ; ) {
        b = tt [ c = a ] ;
        while ( b != lownod ) {
                 if ( !nodfork [ b ] ) 
                     lef [ b ] = lassis [ b ] = c ;
                 else 
                     sis [ sis [ c ] = lassis [ b ] ] = rig [ b ] = c ;
                 if ( ++ nodfork [ b ] > 1 ) break ;
                 c = b ;
                 b = tt [ b ] ; }}
    return ; 
}         

char downlis[MAXNOD] ;

int dodalen ( void )
{
    int nunos , len , a , nod , c , stats ; 
    nunos = binoptlist ( downlis , anxpt ) ;
    len = 0 ;
    for ( a = nunos ; a -- ; ) {
        nod = downlis [ a ] ;
        for ( c = 0 ; c < nc ; ++ c ) {
            stats = matrix [ lef [ nod ] ] [ c ] & matrix [ rig [ nod ] ] [ c ] ;
            if ( !stats ) {
               stats = matrix [ lef [ nod ] ] [ c ] | matrix [ rig [ nod ] ] [ c ] ;
               ++ len ; }
            matrix [ nod ] [ c ] = stats ; }}
    return len ; 
}    

double treefac , divfac [ 50 ] ;   // paths should be MUCH shorter than this, but just in case...
int atlevel ;
int from [ 50 ] ;
Treetyp * curtree ; 

void savant ( what )
{
    int a ;
    fprintf ( stdout , "(" ) ;
    for ( a = 0 ; a < nodz ; ++ a )
      if ( anxpt [ a ] == what )
        if ( a >= nt ) savant ( a ) ;
        else fprintf ( stdout , "%c " , taxname[a] ) ;
    fprintf ( stdout , ")" ) ;
}

void main ( int argc , char ** argv )
{
    int a ;
    char * fnameis ; 
    if ( argc < 2 ) {
        fprintf ( stderr , 
        "\n"
        "  WAGBIAS\n"
        "\n"
        "  Calculates (by enumeration) probabilities of Wagner\n"
        "  trees\n"
        "\n"
        "  Usage:\n"
        "\n"
        "     wagbias [up,down,asis] datafile [ab] [cd] [etc] [ab/cd]\n"
        "\n"
        "  datafile is in xread format, but without xread itself; taxon\n"
        "  names must be single letters. Only states 0-7 allowed.\n"
        "\n"
        "  \"up\" or \"down\" indicate up or down insertion sequence\n"
        "  (otherwise, insertion sequence is fully random).\n"
        "\n"
        "  \"asis\" indicates a single (given) addition sequence\n"
        "\n"
        "  \"ab\", \"cd\", etc., are groups to test (probability is the\n"
        "  sum of probabilities of all Wagner trees with the group)\n"
        "\n"
        "  \"ab/cd is the probability of the partition to be AB|CD\n"
        "\n"
        "  You can redirect output with \"> outfile\"\n"
        "\n"
         ) ; 
        exit ( 0 ) ; }
    fprintf ( stdout , "\n\nOptions: " ) ; 
    for ( a = 1 ; a < argc ; ++ a ) fprintf ( stdout , "%s " , argv[a] ) ;
    fprintf ( stdout , "\n" ) ;
    fflush ( stdout ) ; 
    a = 1 ;
    if ( !strcmp ( argv[1] , "asis" ) ) {
        asisonly = 1 ;
        a = 2 ; }
    else 
    if ( !strcmp ( argv[1] , "up" ) ) {
        upinserts = 1 ;
        a = 2 ; }
    else 
    if ( !strcmp ( argv[1] , "down" ) ) {
        downinserts = 1 ;
        a = 2 ; }
    if ( a >= argc ) {
        fprintf ( stderr , "No data!" ) ;
        exit ( 0 ) ; }
    readdata ( argv [ a ] ) ;
    if ( a + 1 < argc ) 
          while ( ++ a < argc ) 
            addtestgroup ( argv [ a ] ) ; 
    totperms = 1 ;
    for ( a = 1 ; a < nt ; ++ a ) totperms *= a ;
    totperms /= 2 ;  // because XAB is the same as XBA... 
    permlist [ 0 ] = 0 ; 
    stage = 1 ; 
    for ( a = 0 ; a < MAXTREES ; ++ a ) trees [ a ] . prob = 0 ;
    if ( asisonly ) totperms = 1 ; 
    allorders () ;
    fprintf ( stderr , "\rDone!                                        \n" ) ; 
    showresults () ; 
    do_user_groups () ;
    exit ( 1 ) ; 
}

void allorders ( void )
{
    int a , b , isin ;
    if ( stage > 2 )
         if ( permlist [ 1 ] > permlist [ 2 ] ) return ; // only do XAB, never XBA
    for ( a = 0 ; a < nt ; ++ a ) {
        for ( b = isin = 0 ; b < stage && !isin ; ++ b )
           if ( permlist [ b ] == a ) isin = 1 ; 
        if ( isin ) continue ; 
        permlist [ stage ++ ] = a ;
        if ( stage == nt ) 
            doasequence ( ) ;
        else allorders ( ) ;
        if ( asisonly && numperms ) break ; 
        -- stage ; }
}

void doasequence ( void )
{
    int a ;
    double myprop ;
    static int rounds = 0 ; 
    ++ numperms ;
    if ( rounds ++ == 127 ) {
       myprop = ( double ) numperms / totperms ;
       myprop *= 100 ; 
       fprintf ( stderr , "\rWorking ... %.2f%%" , myprop ) ;
       rounds = 0 ; }
    anxpt = anx ; 
    anx [ 0 ] = anx [ nt + 1 ] = nt;
    anx [ permlist [ 1 ] ] = anx [ permlist [ 2 ] ] = nt + 1;
    anx [ nt ] = nt + 2 ; 
    lef [ nt ] = 0 ;
    rig [ nt ] = sis [ 0 ] = nt + 1;
    lef [ nt + 1 ] = permlist [ 1 ] ;
    rig [ nt + 1 ] = sis [ permlist [ 1 ] ] = permlist [ 2 ] ;
    sis [ nt + 1 ] = 0; 
    sis [ permlist [ 2 ] ] = permlist [ 1 ] ;
    htuz = nt + 2 ;
    listhas = agre = 3;
    totax = nt ;
    curprobis = ( double ) 1 / totperms ; 
    addtaxa ( ) ;
}    

char poslist[ MAXTAX ] [ MAXNOD ] ;
int curbestscore ;
char nulist[ MAXNOD ] ; 

void addtaxa ( void )
{  int nowadd = permlist [ agre ] ;
   int i ; 
   int npositions = 0 , bestscore = 10000000 , thislen ;
   char * mypos = poslist [ agre ] ;
   double curprobwas = curprobis ;
   int dirnpos , dist , bestdist , at ; 
   Treetyp * entry , * was ; 
   if ( agre == nt ) {
        for ( i = 0 ; i < nodz ; ++ i ) 
             tmptree [ i ] = anxpt [ i ] ;
        tmptree [ nt ] = nodz ; 
        reset_nodz ( tmptree ) ; 
        hashtree ( tmptree ) ;
        was = entry = ( Treetyp * ) treelist [ curlokey ] ;
        if ( entry != NULL ) {
           while ( entry != NULL ) {
              if ( entry -> key == curkey ) break ; 
              entry = ( Treetyp * ) entry -> to ; }}
        if ( entry == NULL ) {
            if ( ntrees >= MAXTREES ) {
                fprintf ( stderr , "\nOOPS... too many trees found... increase MAXTREES and retry\n(press <enter> to exit..." ) ;
                getc ( stdin ) ;  
                exit ( 0 ) ; }
            entry = treelist [ curlokey ] = trees + ntrees ;
            for ( i = 0 ; i < nodz ; ++ i ) 
               entry -> tree [ i ] = tmptree [ i ] ;
            entry -> key = curkey ;
            entry -> lokey = curlokey ; 
            entry -> prob = 0 ;
            entry -> len = curbestscore ;
            entry -> to = was ; 
            ++ ntrees ; }
         entry -> prob += curprobis ;
         return ; }
   // zert at terms
   anxpt [ nowadd ] = htuz ; 
   for ( i = 1 ; i < agre ; ++ i ) {
        zert ( nowadd , permlist [ i ] ) ;
        thislen = dodalen ( ) ;
        unzert ( nowadd ) ;
        if ( thislen < bestscore ) {
            npositions = 0 ;
            bestscore = thislen ; }
        if ( thislen == bestscore ) 
            mypos [ npositions ++ ] = permlist [ i ] ; }
   // zert at intnodes 
   for ( i = nt + 1 ; i < htuz ; ++ i ) {
        zert ( nowadd , i ) ;
        thislen = dodalen ( ) ;
        unzert ( nowadd ) ;
        if ( thislen < bestscore ) {
            npositions = 0 ;
            bestscore = thislen ; }
        if ( thislen == bestscore ) 
            mypos [ npositions ++ ] = i ; }
    // now, use all best positions...         
    ++ agre ;
    ++ htuz ;
    if ( agre == nt ) {
       curbestscore = bestscore ; 
       if ( bestscore < mptlen ) mptlen = bestscore ; }
    if ( upinserts || downinserts ) {
       if ( upinserts ) bestdist = 10000 ;
       else bestdist = 0 ; 
       for ( i = 0 ; i < npositions ; ++ i ) {
           for ( at = mypos [ i ] , dist = 0 ; at != nt ; at = anxpt [ at ] ) ++ dist ;
           if ( dist < bestdist && upinserts ) {
               dirnpos = 0 ;
               bestdist = dist ; }
           if ( dist > bestdist && downinserts ) {
               dirnpos = 0 ;
               bestdist = dist ; }
           if ( dist == bestdist )
               nulist [ dirnpos ++ ] = mypos [ i ] ; }
       for ( i = 0 ; i < dirnpos ; ++ i )
           mypos [ i ] = nulist [ i ] ;
       npositions = dirnpos ; }
    curprobis = curprobis / npositions ;
    for ( i = 0 ; i < npositions ; ++ i ) {
        zert ( nowadd , mypos [ i ] ) ; 
        addtaxa () ;
        unzert ( nowadd ) ; }
    curprobis = curprobwas ; 
    -- agre ;
    -- htuz ; 
}

void showresults ( void )
{
    int i ;
    Treetyp * myt ;
    double probsum = 0 ;
    int nummpts = 0 , numnonmpts = 0 ; 
    fprintf ( stdout , "\n\n" ) ; 
    for ( i = 0 , myt = trees ; i < ntrees ; ++ i , ++ myt ) {
       if ( myt -> len == mptlen ) {
            fprintf ( stdout , "TREE %4i " , i ) ;
            ++ nummpts ; }
       else {
           fprintf ( stdout , "tree %4i " , i ) ;
           ++ numnonmpts ; }
       anxpt = myt -> tree ;
       savant ( nt ) ;
       fprintf ( stdout , " ; L = %i - P = %.12f\n" , myt -> len , myt -> prob ) ;
       if ( myt -> prob > damaxprob ) damaxprob = myt -> prob ; 
       if ( myt -> prob < daminprob ) daminprob = myt -> prob ; 
       probsum += myt -> prob ; }
       fprintf ( stdout , "\nTOTAL PROB.: %.8f (min-max %.6f - %.6f, ratio %.4f)\n\n" , probsum , daminprob , damaxprob , damaxprob / daminprob ) ;
       fprintf ( stdout , "%i trees of L=%i, and %i suboptimal trees\n" , nummpts , mptlen , numnonmpts ) ; 
}

void do_user_groups ( void )
{
    double totfreq[25] ;
    int a , i , n , j , k , isamatch , nod , nunos ;
    char list[20] ;
    double eqfactor ;
    unsigned long int myinner , myouter ; 
    Treetyp * myt ; 
    for ( a = 0 ; a < nt ; ++ a )
       testmat [ a ] = 1 << a ;
    if ( numtestpartits ) {
        for ( a = 0 ; a < numtestpartits ; ++ a ) 
            totfreq [ a ] = 0 ;
        for ( a = 0 , myt = trees ; a < ntrees ; ++ a , ++ myt ) {
            anxpt = myt -> tree ;
            binmksis ( anxpt ) ;
            nunos = binoptlist ( list , anxpt ) ;
            for ( i = nunos ; i -- ; ) {
                nod = list [ i ] ;
                testmat [ nod ] = testmat [ lef [ nod ] ] | testmat [ rig [ nod ] ] ; }
            for ( n = 0 ; n < numtestpartits ; ++ n ) {
                isamatch = 0 ;
                myinner = myouter = 0 ;
                for ( k = 0 ; k < partitsize [ n ] ; ++ k )
                  if ( k < partitat [ n ] ) 
                      myouter |= ( 1 << testpartit [ n ] [ k ] ) ;
                  else 
                      myinner |= ( 1 << testpartit [ n ] [ k ] ) ;
                for ( i = nunos ; i -- && !isamatch ; ) {
                    nod = list [ i ] ;
                    if ( ( myinner & testmat [ nod ] ) == myinner && !( testmat [ nod ] & myouter ) ) isamatch = 1 ; 
                    if ( ( myouter & testmat [ nod ] ) == myouter && !( testmat [ nod ] & myinner ) ) isamatch = 1 ; }
                if ( isamatch ) 
                    totfreq [ n ] += myt -> prob ; }}
        fprintf ( stdout , "\nPARTITION SUMMARY: \n\n       partition       TOTAL         expected\n\n   " ) ;
        for ( a = 0 ; a < numtestpartits ; ++ a ) {
            fprintf ( stdout , "    " ) ; 
            for ( k = 0 ; k < partitsize [ a ] ; ++ k ) {
                if ( k == partitat [ a ] ) 
                     fprintf ( stdout , "|"  ) ; 
                fprintf ( stdout , "%c" , taxname [ testpartit [ a ] [ k ] ] ) ; }
            for ( k = partitsize [ a ] ; k < 10 ; ++ k ) fprintf ( stdout , " " ) ;     
            fprintf ( stdout , "     %2.8f\n   " , totfreq [ a ] ) ; }
        fprintf ( stdout , "\n" ) ; }
    if ( !numtestgroups ) return ;
    for ( a = 0 ; a < numtestgroups ; ++ a ) 
        totfreq [ a ] = 0 ;
    for ( a = 0 , myt = trees ; a < ntrees ; ++ a , ++ myt ) {
        anxpt = myt -> tree ;
        binmksis ( anxpt ) ;
        nunos = binoptlist ( list , anxpt ) ;
        for ( i = nunos ; i -- ; ) {
            nod = list [ i ] ;
            testmat [ nod ] = testmat [ lef [ nod ] ] | testmat [ rig [ nod ] ] ; }
        for ( n = 0 ; n < numtestgroups ; ++ n ) {
            isamatch = 0 ; 
            for ( j = nt + 1 ; j < nodz && !isamatch ; ++ j )
                if ( testmat [ j ] == testgroup [ n ] ) isamatch = 1 ; 
            if ( isamatch ) 
                totfreq [ n ] += myt -> prob ; }}
    fprintf ( stdout , "\nGROUP SUMMARY: \n\n         group     FREQ\n\n   " ) ;
    for ( a = 0 ; a < numtestgroups ; ++ a ) {
        for ( j = testgroupsz[ a ] ; j < 10 ; ++ j ) fprintf ( stdout , " " ) ; 
        for ( i = 0 ; i < nt ; ++ i ) 
             if ( ( ( 1 << i ) & testgroup [ a ] ) )
                 fprintf ( stdout , "%c" , toupper ( taxname[i] ) ) ; 
        fprintf ( stdout , "      %2.8f\n   " , totfreq[a] ) ; }
    fprintf ( stdout , "\n" ) ; 
}




